from asyncio import create_task
from importlib.metadata import entry_points
import logging

from gisaf.live import live_server
from gisaf.redis_tools import Store

logger = logging.getLogger('Gisaf admin manager')

class AdminManager:
    """
    Application wide manager of the admin (baskets).
    One instance only, handled by Gisaf's process.
    """
    store: Store
    async def setup_admin(self, app):
        """
        Create the default baskets, scan and create baskets
        from the Python entry points.
        Runs at startup.
        """
        self.app = app
        self.store = app['store']

        ## Standard baskets
        from gisaf.baskets import Basket, standard_baskets
        self.baskets = {
            basket.name: basket
            for basket in standard_baskets
        }

        for entry_point in entry_points().select(group='gisaf_extras.baskets'):
            try:
                basket_class = entry_point.load()
            except ModuleNotFoundError as err:
                logger.warning(err)
                continue
            if issubclass(basket_class, Basket):
                ## Get name, validity check
                if basket_class.name == None:
                    name = entry_point.name
                else:
                    name = basket_class.name
                if name in self.baskets:
                    logger.warn(f'Skip basket {name} in {entry_point.module}: name already defined')
                    continue
                ## Instanciate
                basket = basket_class()
                basket._custom_module = entry_point.name
                ## Check base_dir, eventually create it
                if not basket.base_dir.exists():
                    try:
                        basket.base_dir.mkdir()
                    except Exception as err:
                        logger.warn(f'Skip basket {name} in {entry_point.module}: '
                                    f'cannot create directory for basket {name} at {basket.base_dir}')
                        continue
                    else:
                        logger.info(f'Created directory for basket {name} at {basket.base_dir}')
                ## Add to register
                self.baskets[name] = basket
                logger.info(f'Added Basket {entry_point.name} from {entry_point.module}')

        ## Give a reference to the application to the baskets
        for basket in self.baskets.values():
            basket.app = app

        ## Subscribe to admin redis channels
        self.pub_categories = self.store.redis.pubsub()
        self.pub_scheduler = self.store.redis.pubsub()
        await self.pub_categories.psubscribe('admin:categories:update')
        task1 = create_task(self._listen_to_redis_categories())
        await self.pub_scheduler.psubscribe('admin:scheduler:json')
        task2 = create_task(self._listen_to_redis_scheduler())

        app['admin'] = self

    async def baskets_for_role(self, request):
        return {
            name: basket for name, basket in self.baskets.items()
            if await basket.allowed_for(request)
        }

    async def _listen_to_redis_categories(self):
        """
        Subscribe the redis sub channel for category updates ("admin:categories:update")
        """
        async for msg in self.pub_categories.listen():
            if msg['type'] == 'pmessage':
                ## XXX: Why the name isn't retrieved?
                #client = await self.app['store'].pub.client_getname()
                client = self.app['store'].uuid

                ## !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                ## FIXME: pubsub admin:categories:update
                ## !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

                ## Skip for the process which sent this message actually updated its registry
                #breakpoint()
                if client != msg['data'].decode():
                    from gisaf.database import make_auto_models
                    await make_auto_models(self.app)

    async def _listen_to_redis_scheduler(self):
        """
        Subscribe the redis sub channel for scheduler jobs ("admin:scheduler:json")
        """
        async for msg in self.pub_scheduler.listen():
            if msg['type'] == 'pmessage':
                await live_server._send_to_ws_clients(msg['channel'].decode(), msg['data'].decode())


manager = AdminManager()